#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from .pkg.loggerFactory import StreamOnlyLogger
except:
    from pytq.pkg.loggerFactory import StreamOnlyLogger


class ClassWithLogger(object):
    """
    A class with a built-in logger.
    """

    def __init__(self, logger=None, **kwargs):
        if logger is None:
            self.logger = StreamOnlyLogger()
        else:
            self.logger = logger
        self.verbose = True

    def log_on(self):
        """
        Turn on logger.
        """
        self.verbose = True
        self.logger.enable_verbose = True

    def log_off(self):
        """
        Turn off logger.
        """
        self.verbose = False
        self.logger.enable_verbose = False

    def debug(self, msg, indent=0):
        """
        Log a debug message.
        """
        if self.verbose:
            self.logger.debug(msg, indent)

    def info(self, msg, indent=0):
        """
        Log a info message.
        """
        if self.verbose:
            self.logger.info(msg, indent)

    def warning(self, msg, indent=0):
        """
        Log a warning message.
        """
        if self.verbose:
            self.logger.warning(msg, indent)

    def error(self, msg, indent=0):
        """
        Log a error message.
        """
        if self.verbose:
            self.logger.error(msg, indent)

    def critical(self, msg, indent=0):
        """
        Log a critical message.
        """
        if self.verbose:
            self.logger.critical(msg, indent)

    def show(self, msg, indent=0):
        """
        Print to console.
        """
        self.logger.show(msg, indent=indent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import types
import pickle
from multiprocessing import cpu_count
from multiprocessing.dummy import Pool

try:
    from .task import Task
    from .per_class_logger import ClassWithLogger
    from .pkg.fingerprint import hash_data
    from .pkg.exception_mate import get_last_exc_info
except:  # pragma: no cover
    from pytq.task import Task
    from pytq.per_class_logger import ClassWithLogger
    from pytq.pkg.fingerprint import hash_data
    from pytq.pkg.exception_mate import get_last_exc_info


class BaseScheduler(ClassWithLogger):
    """
    All Scheduler has to inherit from this base class.

    Step1. Workflow:

    Generate task queue, it is a list of input_data.

    Step2. Pre-process input data:

    1. remove duplicate.
    2. generate :class:`~pytq.task.Task`.

    :meth:`~BaseScheduler._default_batch_pre_process` method will be called by
    default.

    Step3. For each task, :meth:`~BaseScheduler._process` method will be called,
    includes:

    1. pre_process
    2. user_process, input_data -> output_data
    3. post_process
    """

    def __init__(self, logger=None):
        super(BaseScheduler, self).__init__(logger=logger)

        # link hash input method
        try:
            self.user_hash_input(None)
            self._hash_input = self.user_hash_input
        except NotImplementedError:
            self._hash_input = self._default_hash_input
        except:
            self._hash_input = self.user_hash_input

        # link duplicate test method
        try:
            self.user_is_duplicate(None)
            self._is_duplicate = self.user_is_duplicate
        except NotImplementedError:
            self._is_duplicate = self._default_is_duplicate
        except:
            self._is_duplicate = self.user_is_duplicate

        # link post process method
        try:
            self.user_batch_pre_process(None)
            self._batch_pre_process = self.user_batch_pre_process
        except NotImplementedError:
            self._batch_pre_process = self._default_batch_pre_process
        except:
            self._batch_pre_process = self.user_batch_pre_process

        # link pre process method
        try:
            self.user_pre_process(None)
            self._pre_process = self.user_pre_process
        except NotImplementedError:
            self._pre_process = self._default_pre_process
        except:
            self._pre_process = self.user_pre_process

        # link post process method
        try:
            self.user_post_process(None)
            self._post_process = self.user_post_process
        except NotImplementedError:
            self._post_process = self._default_post_process
        except:
            self._post_process = self.user_post_process

        # if True, error will not stop the program
        # if False, error will be immediately raised
        self.ignore_error = None

    def _default_hash_input(self, input_data):
        """
        Default hash method to get a identical fingerprint for input data.

        By default its: pickle the data and md5 it.

        This method will be used when :meth:`BaseScheduler.user_hash_input`
        are not defined.

        :returns: fingerprint for ``input_data``
        :rtype: string or integer. depends on use case.

        **中文文档**

        默认的 取 ``input_data`` 指纹的操作。
        实际上是对 ``pickle.dumps`` 后的数据取 md5 指纹。
        """
        return hash_data(input_data)

    def user_hash_input(self, input_data):
        """
        (Optional) Get identical fingerprint for input data.

        :returns: fingerprint for ``input_data``
        :rtype: string or integer. depends on use case.

        **中文文档**

        (可选自定义) 用户自己定义的取 ``input_data`` 指纹的操作。如果不定义，则使用

        :meth:`BaseScheduler._default_hash_input` 方法。
        """
        raise NotImplementedError

    def _hash_input(self, input_data):
        """
        The real hashing method will be called.

        :returns: fingerprint for ``input_data``
        :rtype: string or integer. depends on use case.

        **中文文档**

        实际被调用的取指纹方法。
        """
        raise NotImplementedError

    def _default_is_duplicate(self, task):
        """
        Default duplicate test method, always not duplicate.

        :return: return True, when it's a duplicate item.
        :rtype: boolean.

        **中文文档**

        默认的任务排重检测。实际上是全部视为 ``不重复``。
        """
        return False  # Alwasy not duplicate

    def user_is_duplicate(self, task):
        """
        (Optional) Check if a task is duplicate.

        :return: return True, when it's a duplicate item.
        :rtype: boolean.

        .. warning::

            If you customized this method, usually you also need to implement
            :meth:`~BaseScheduler.user_batch_pre_process` method.

            Because default batch pre-process includes duplicate filter.

        **中文文档**

        (可选自定义) 用户自定义的任务排重检测。
        """
        raise NotImplementedError

    def _is_duplicate(self, task):
        """
        The real duplicate test method will be called.

        :return: return True, when it's a duplicate item.
        :rtype: boolean.

        **中文文档**

        (可选自定义) 用户自定义的任务排重检测。
        """
        raise NotImplementedError

    def _remove_duplicate(self, input_data_queue):
        """
        Remove duplicate input_data. And pack ``input_data`` to
        :class:`pytq.task.Task`.

        This method will be used for default batch pre-process when
        ``pre_process`` are not given in :meth:`~BaseScheduler.do`.

        :param input_data_queue:
        :returns: task_queue.
        :rtype: types.GeneratorType.

        **中文文档**

        移除那些重复的输入数据，并将 ``input_data`` 打包成 :class:`pytq.task.Task`。
        """
        is_generator = isinstance(input_data_queue, types.GeneratorType)

        if not is_generator:
            left_counter = len(input_data_queue)

        nth_counter = 0
        for input_data in input_data_queue:
            if is_generator:
                left_counter = None
            else:
                left_counter -= 1

            task = Task(
                id=self._hash_input(input_data),
                input_data=input_data,
                nth_counter=nth_counter,
                left_counter=left_counter,
            )

            if not self._is_duplicate(task):
                nth_counter += 1
                yield task

    def user_batch_pre_process(self, input_data_queue):
        """
        A method will be called to pre process task queue before doing any real
        per task process. Usually it can be duplicate filter, statistic check.

        :param input_data_queue:
        :return: task_queue, iterable object, item in it has to be
        :class:`~pytq.task.Task`. Recommend to implement ``task.nth_counter``
        and ``task.left_counter`` variable.
        """
        raise NotImplementedError

    _default_batch_pre_process = _remove_duplicate
    _default_batch_pre_process.__doc__ = _remove_duplicate.__doc__

    def _batch_pre_process(self, input_data_queue):
        """
        The real method will be called for batch pre-process.
        """
        raise NotImplementedError

    def user_pre_process(self, task):
        """
        (Optional) Defines the action that before the
        :meth:`BaseScheduler.user_process() been called.
        Will be called when :attr:`pytq.task.Task.pre_process` are not defined.

        :param task: :class:`pytq.task.Task` instance.
        """
        raise NotImplementedError

    def _default_pre_process(self, task):
        """
        Default behavior of user_pre_process, do nothing.
        """
        pass

    def _pre_process(self, task):
        """
        The real method will be called for pre_process.
        """
        raise NotImplementedError

    def user_post_process(self, task):
        """
        (Optional) Defines the action that after the
        :meth:`~BaseScheduler.user_process() been called.
        Will be called when :attr:`pytq.task.Task.post_process` are not defined.

        .. warning::

            When you customized this method, usually you also need to update
            :meth:`~BaseScheduler.get` method. Because post process usually
            is used to write output_data to data persistence layer. If you
            changed the way you store it, you have to change the way you read
            it.

        :param task: :class:`pytq.task.Task` instance.
        """
        raise NotImplementedError

    def _default_post_process(self, task):
        """
        Default behavior of user_post_process, do nothing.
        """
        pass

    def _post_process(self, task):
        """
        The real method will be called for post_process.
        """
        raise NotImplementedError

    def user_process(self, input_data):
        """
        (Required) Defines the logic that process the input_data, returns
        output_data

        :param input_data:
        :return: the output_data
        """
        raise NotImplementedError

    def _process_one(self, task):
        """
        Process one task.

        **中文文档**

        处理数据。包含三个步骤：

        1. 预处理
        2. 处理
        3. 后处理
        """
        # if task.pre_process not given, use scheduler._pre_process
        if task.pre_process is None:
            self._pre_process(task)
        # if given, use task.pre_process
        else:
            task._pre_process()

        output_data = self.user_process(task.input_data)
        task.output_data = output_data

        # if task.post_process not given, use scheduler._post_process
        if task.post_process is None:
            self._post_process(task)
        # if given, use task.post_process
        else:
            task._post_process()

    def _process(self, task):
        """
        The real processing method will be called.
        """
        self.info(task.progress_msg())

        if self.ignore_error:
            try:
                self._process_one(task)
                self.info("Success!", 1)
            except Exception as e:
                self.info("Failed due to: %r" % get_last_exc_info(), 1)
        else:
            self._process_one(task)
            self.info("Success!", 1)

    def _do_single_process(self, task_queue):
        """
        Execute single thread process.

        :param task_queue: task queue/list.
        """
        for task in task_queue:
            self._process(task)

    def _do_multi_process(self, task_queue, processes=None):
        """
        Execute multi thread process.

        :param task_queue: task queue/list.
        """
        if processes is None:
            processes = cpu_count()
        pool = Pool(processes=processes)
        pool.map(self._process, task_queue)

    def do(self,
           input_data_queue,
           pre_process=None,
           multiprocess=False,
           processes=None,
           ignore_error=True):
        """
        Process all input_data.

        :param input_data_queue: list of input data (or generator).
        :param pre_process: a callable function take input_data_queue, and
          pre-process it, returns a task_queue (iterable object, item are
          :class:`~pytq.task.Task`.
        :param multiprocess: trigger to use multiprocess.

        **中文文档**

        处理数据序列中的所有数据。

        1. 预处理所有数据，将其打包成 ``task_queue``。
        2. 进行单线程处理或是多线程处理。
        """
        if pre_process is None:
            task_queue = self._batch_pre_process(input_data_queue)
        else:
            task_queue = pre_process(input_data_queue)

        try:
            self.info("\nHas %s items todo." % len(input_data_queue))
        except:
            self.info("\nHas UNKNOWN items todo")

        self.ignore_error = ignore_error

        if multiprocess:
            self._do_multi_process(task_queue, processes=processes)
        else:
            self._do_single_process(task_queue)

        self.info("Complete!")

    def clear_all(self):
        """
        Clear all data.

        **中文文档**

        重置Filter至初始状态。
        """
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def get(self, id):
        """
        Get output data by fingerprint of input_data.

        **中文文档**

        根据输入的指纹, 直接获得已经完成的输出数据。
        """
        raise NotImplementedError

    def get_output(self, input_data):
        """
        Get output data of the input_data.

        **中文文档**

        根据输入的数据, 直接获得已经完成的输出的数据。
        """
        return self.get(self._hash_input(input_data))

    def keys(self):
        return iter(self)

    def items(self):
        for key in self:
            yield (key, self.get(key))


class BaseDBTableBackedScheduler(BaseScheduler):
    """
    Scheduler that use database table as backend storage.

    - Task.id as primary_key
    - Other column / field for input_data, output_data storage.
    """

    def _get_finished_id_set(self):
        """
        A method that returns all saved id set.
        (For all processed input_data)

        :returns: a id set.
        :rtype: set.
        """
        raise NotImplementedError

    def _default_batch_pre_process(self, input_data_queue):
        """
        Default method to pre-process input_data_queue in Database Table backed
        scheduler. The logic is:

        1. get all finished _id set.
        2. filter out all input_data that fingerprint falls in that set.
        """
        finished_id_set = self._get_finished_id_set()

        is_generator = isinstance(input_data_queue, types.GeneratorType)

        # if its generator, then left_counter always be None
        if not is_generator:
            left_counter = len(input_data_queue)

        nth_counter = 0
        for input_data in input_data_queue:
            if is_generator:
                left_counter = None
            else:
                left_counter -= 1

            id = self._hash_input(input_data)
            if id not in finished_id_set:
                nth_counter += 1
                task = Task(
                    id=id,
                    input_data=input_data,
                    nth_counter=nth_counter,
                    left_counter=left_counter,
                )
                yield task


class Encoder(object):
    def link_encode_method(self):
        """
        Bind encode method.
        """
        # link encode method
        try:
            self.user_encode(None)
            self._encode = self.user_encode
        except NotImplementedError:
            self._encode = self._default_encode
        except:
            self._encode = self.user_encode

        # link decode method
        try:
            self.user_decode(None)
            self._decode = self.user_decode
        except NotImplementedError:
            self._decode = self._default_decode
        except:
            self._decode = self.user_decode

    def _default_encode(self, obj):
        return pickle.dumps(obj)

    def user_encode(self, obj):
        """
        (Optional) User defined serializer for output_data.

        :returns: bytes or string.

        **中文文档**

        用于对处理结果序列化的函数。默认使用pickle。
        """
        raise NotImplementedError

    def _encode(self, obj):
        raise NotImplementedError

    def _default_decode(self, bytes_or_str):
        return pickle.loads(bytes_or_str)

    def user_decode(self, bytes_or_str):
        """
        (Optional) User defined deserializer for output_data.

        :returns: python object.

        **中文文档**

        用于对处理结果反序列化的函数。默认使用pickle。
        """
        raise NotImplementedError

    def _decode(self, bytes_or_str):
        raise NotImplementedError


class StatusFlag(object):
    """
    MongoDB collection backed scheduler.

    Feature:

    1. there's pre-defined integer - ``duplicate_flag``, will be stored in
        ``status`` column / field.
    2. there's a ``edit_at`` datetime field, represent the
        last time the document been edited.

    .. note::

        Any value greater or equal than ``duplicate_flag``, AND the ``edit_at``
        time is smaller ``update_interval`` seconds ago, means it is a duplicate
        item.

    :param duplicate_flag: int, represent a status code for finished / duplicate
    :param update_interval: int, represent need-to-update interval (unit: seconds)
    :param status_key: str.
    :param edit_key: str.
    """

    duplicate_flag = None  # integer
    """
    A integer value represent its a duplicate item. Any value greater or equal 
    than this will be a duplicate item, otherwise its not.

    You could define that when you initiate the scheduler.
    """

    update_interval = None  # integer, represent xxx seconds
    """
    If a item has been finished more than ``update_interval`` seconds, then
    it should be re-do, and it is NOT a duplicate item.

    You could define that when you initiate the scheduler. 
    """

    status_key = "_status"
    edit_at_key = "_edit_at"

    def pre_process_duplicate_flag_and_update_interval(self,
                                                       duplicate_flag,
                                                       update_interval):
        """
        bind settings.
        """
        if duplicate_flag is not None:
            self.duplicate_flag = duplicate_flag
        if update_interval is not None:
            self.update_interval = update_interval

    @property
    def duplicate_flag(self):
        """
        A integer value represent its a duplicate item. Any value greater or equal
        than this will be a duplicate item, otherwise its not.

        You could define that when you initiate the scheduler.
        """
        raise NotImplementedError

    @property
    def update_interval(self):
        """
        If a item has been finished more than ``update_interval`` seconds, then
        it should be re-do, and it is NOT a duplicate item.

        You could define that when you initiate the scheduler.
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import attr
import collections


class AttrsClass(object):
    def keys(self):
        return [a.name for a in self.__attrs_attrs__]

    def values(self):
        return [getattr(self, a.name) for a in self.__attrs_attrs__]

    def items(self):
        return list(zip(self.keys(), self.values()))

    def to_dict(self):
        return attr.asdict(self)

    def to_OrderedDict(self):
        return collections.OrderedDict(self.items())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pymongo import collection as collection_module
try:
    from .scheduler import Task, BaseDBTableBackedScheduler, Encoder
except:  # pragma: no cover
    from pytq.scheduler import Task, BaseDBTableBackedScheduler, Encoder


class MongoDBScheduler(BaseDBTableBackedScheduler, Encoder):
    """
    MongoDB collection backed scheduler.

    Feature:

    1. fingerprint of :meth:`~MongoDBScheduler._hash_input()` will be ``_id``
      field in MongoDB collection.
    2. output_data will be serialized and stored in ``out`` field.

    :param collection: :class:`pymongo.Collection`.
    """
    collection = None
    """
    Backend :class:`pymongo.Collection`. You could define that when you
    initiate the scheduler.
    """

    output_key = "_out"
    """
    Default field name used to store output_data
    """

    def __init__(self, logger=None, collection=None):
        super(MongoDBScheduler, self).__init__(logger=logger)
        self.prepare_collection(collection)
        self.link_encode_method()

    def prepare_collection(self, collection):
        if collection is not None:
            self.collection = collection
        if self.collection is None:
            raise NotImplementedError("Please specify a mongodb collection!")
        self.col = self.collection

    @property
    def _col(self):
        import warnings
        warnings.warn(
            DeprecationWarning("this property will be deprecated soon!"))
        return self.collection

    def _default_is_duplicate(self, task):
        """
        Check if ``task.id`` presents in the collection.
        """
        return self.col.find_one({"_id": task.id}) is not None

    def _get_finished_id_set(self):
        """
        It's Primary key value set.
        """
        return set([
            doc["_id"] for doc in self.col.find({}, {"_id": True})
        ])

    def _default_post_process(self, task):
        """
        Save output_data into ``out`` field.
        """
        self.col.update(
            {"_id": task.id},
            {"$set": {self.output_key: self._encode(task.output_data)}},
            upsert=True,
        )

    def __len__(self):
        return self.col.find().count()

    def __iter__(self):
        for doc in self.col.find():
            yield doc["_id"]

    def clear_all(self):
        self.col.remove({})

    def get_output(self, input_data):
        key = self.user_hash_input(input_data)
        return self._decode(self.col.find_one({"_id": key})[self.output_key])

    def items(self):
        for doc in self.col.find():
            yield (doc["_id"], self._decode(doc[self.output_key]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pickle
import hashlib


def hash_data(data):
    """
    Fingerprint of the data.
    """
    m = hashlib.md5()
    m.update(pickle.dumps(data))
    return m.hexdigest()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This code is distributed under the terms and conditions
# from the Apache License, Version 2.0
#
# http://opensource.org/licenses/apache2.0.php
#
# This code was inspired by:
#  * http://code.activestate.com/recipes/576638-draft-for-an-sqlite3-based-dbm/
#  * http://code.activestate.com/recipes/526618/

"""
A lightweight wrapper around Python's sqlite3 database, with a dict-like interface
and multi-thread access support::

>>> mydict = SqliteDict('some.db', autocommit=True) # the mapping will be persisted to file `some.db`
>>> mydict['some_key'] = any_picklable_object
>>> print mydict['some_key']
>>> print len(mydict) # etc... all dict functions work

Pickle is used internally to serialize the values. Keys are strings.

If you don't use autocommit (default is no autocommit for performance), then
don't forget to call `mydict.commit()` when done with a transaction.

"""

import sqlite3
import os
import sys
import tempfile
import random
import logging
import traceback

from threading import Thread

try:
    __version__ = __import__(
        'pkg_resources').get_distribution('sqlitedict').version
except:
    __version__ = '?'

major_version = sys.version_info[0]
if major_version < 3:  # py <= 2.x
    if sys.version_info[1] < 5:  # py <= 2.4
        raise ImportError(
            "sqlitedict requires python 2.5 or higher (python 3.3 or higher supported)")

    # necessary to use exec()_ as this would be a SyntaxError in python3.
    # this is an exact port of six.reraise():
    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")

    exec_("def reraise(tp, value, tb=None):\n"
          "    raise tp, value, tb\n")
else:
    def reraise(tp, value, tb=None):
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

try:
    from cPickle import dumps, loads, HIGHEST_PROTOCOL as PICKLE_PROTOCOL
except ImportError:
    from pickle import dumps, loads, HIGHEST_PROTOCOL as PICKLE_PROTOCOL

# some Python 3 vs 2 imports
try:
    from collections import UserDict as DictClass
except ImportError:
    from UserDict import DictMixin as DictClass

try:
    from queue import Queue
except ImportError:
    from Queue import Queue


logger = logging.getLogger(__name__)


def open(*args, **kwargs):
    """See documentation of the SqliteDict class."""
    return SqliteDict(*args, **kwargs)


def encode(obj):
    """Serialize an object using pickle to a binary format accepted by SQLite."""
    return sqlite3.Binary(dumps(obj, protocol=PICKLE_PROTOCOL))


def decode(obj):
    """Deserialize objects retrieved from SQLite."""
    return loads(bytes(obj))


class SqliteDict(DictClass):
    VALID_FLAGS = ['c', 'r', 'w', 'n']

    def __init__(self, filename=None, tablename='unnamed', flag='c',
                 autocommit=False, journal_mode="DELETE", encode=encode, decode=decode):
        """
        Initialize a thread-safe sqlite-backed dictionary. The dictionary will
        be a table `tablename` in database file `filename`. A single file (=database)
        may contain multiple tables.

        If no `filename` is given, a random file in temp will be used (and deleted
        from temp once the dict is closed/deleted).

        If you enable `autocommit`, changes will be committed after each operation
        (more inefficient but safer). Otherwise, changes are committed on `self.commit()`,
        `self.clear()` and `self.close()`.

        Set `journal_mode` to 'OFF' if you're experiencing sqlite I/O problems
        or if you need performance and don't care about crash-consistency.

        The `flag` parameter. Exactly one of:
          'c': default mode, open for read/write, creating the db/table if necessary.
          'w': open for r/w, but drop `tablename` contents first (start with empty table)
          'r': open as read-only
          'n': create a new database (erasing any existing tables, not just `tablename`!).

        The `encode` and `decode` parameters are used to customize how the values
        are serialized and deserialized.
        The `encode` parameter must be a function that takes a single Python
        object and returns a serialized representation.
        The `decode` function must be a function that takes the serialized
        representation produced by `encode` and returns a deserialized Python
        object.
        The default is to use pickle.

        """
        self.in_temp = filename is None
        if self.in_temp:
            randpart = hex(random.randint(0, 0xffffff))[2:]
            filename = os.path.join(
                tempfile.gettempdir(), 'sqldict' + randpart)

        if flag not in SqliteDict.VALID_FLAGS:
            raise RuntimeError("Unrecognized flag: %s" % flag)
        self.flag = flag

        if flag == 'n':
            if os.path.exists(filename):
                os.remove(filename)

        dirname = os.path.dirname(filename)
        if dirname:
            if not os.path.exists(dirname):
                raise RuntimeError(
                    'Error! The directory does not exist, %s' % dirname)

        self.filename = filename
        if '"' in tablename:
            raise ValueError('Invalid tablename %r' % tablename)
        self.tablename = tablename
        self.autocommit = autocommit
        self.journal_mode = journal_mode
        self.encode = encode
        self.decode = decode

        logger.info("opening Sqlite table %r in %s" % (tablename, filename))
        MAKE_TABLE = 'CREATE TABLE IF NOT EXISTS "%s" (key TEXT PRIMARY KEY, value BLOB)' % self.tablename
        self.conn = self._new_conn()
        self.conn.execute(MAKE_TABLE)
        self.conn.commit()
        if flag == 'w':
            self.clear()

    def _new_conn(self):
        return SqliteMultithread(self.filename, autocommit=self.autocommit, journal_mode=self.journal_mode)

    def __enter__(self):
        if not hasattr(self, 'conn') or self.conn is None:
            self.conn = self._new_conn()
        return self

    def __exit__(self, *exc_info):
        self.close()

    def __str__(self):
        return "SqliteDict(%s)" % (self.filename)

    def __repr__(self):
        return str(self)  # no need of something complex

    def __len__(self):
        # `select count (*)` is super slow in sqlite (does a linear scan!!)
        # As a result, len() is very slow too once the table size grows beyond trivial.
        # We could keep the total count of rows ourselves, by means of triggers,
        # but that seems too complicated and would slow down normal operation
        # (insert/delete etc).
        GET_LEN = 'SELECT COUNT(*) FROM "%s"' % self.tablename
        rows = self.conn.select_one(GET_LEN)[0]
        return rows if rows is not None else 0

    def __bool__(self):
        # No elements is False, otherwise True
        GET_MAX = 'SELECT MAX(ROWID) FROM "%s"' % self.tablename
        m = self.conn.select_one(GET_MAX)[0]
        # Explicit better than implicit and bla bla
        return True if m is not None else False

    def iterkeys(self):
        GET_KEYS = 'SELECT key FROM "%s" ORDER BY rowid' % self.tablename
        for key in self.conn.select(GET_KEYS):
            yield key[0]

    def itervalues(self):
        GET_VALUES = 'SELECT value FROM "%s" ORDER BY rowid' % self.tablename
        for value in self.conn.select(GET_VALUES):
            yield self.decode(value[0])

    def iteritems(self):
        GET_ITEMS = 'SELECT key, value FROM "%s" ORDER BY rowid' % self.tablename
        for key, value in self.conn.select(GET_ITEMS):
            yield key, self.decode(value)

    def keys(self):
        return self.iterkeys() if major_version > 2 else list(self.iterkeys())

    def values(self):
        return self.itervalues() if major_version > 2 else list(self.itervalues())

    def items(self):
        return self.iteritems() if major_version > 2 else list(self.iteritems())

    def __contains__(self, key):
        HAS_ITEM = 'SELECT 1 FROM "%s" WHERE key = ?' % self.tablename
        return self.conn.select_one(HAS_ITEM, (key,)) is not None

    def __getitem__(self, key):
        GET_ITEM = 'SELECT value FROM "%s" WHERE key = ?' % self.tablename
        item = self.conn.select_one(GET_ITEM, (key,))
        if item is None:
            raise KeyError(key)
        return self.decode(item[0])

    def __setitem__(self, key, value):
        if self.flag == 'r':
            raise RuntimeError('Refusing to write to read-only SqliteDict')

        ADD_ITEM = 'REPLACE INTO "%s" (key, value) VALUES (?,?)' % self.tablename
        self.conn.execute(ADD_ITEM, (key, self.encode(value)))

    def __delitem__(self, key):
        if self.flag == 'r':
            raise RuntimeError('Refusing to delete from read-only SqliteDict')

        if key not in self:
            raise KeyError(key)
        DEL_ITEM = 'DELETE FROM "%s" WHERE key = ?' % self.tablename
        self.conn.execute(DEL_ITEM, (key,))

    def update(self, items=(), **kwds):
        if self.flag == 'r':
            raise RuntimeError('Refusing to update read-only SqliteDict')

        try:
            items = items.items()
        except AttributeError:
            pass
        items = [(k, self.encode(v)) for k, v in items]

        UPDATE_ITEMS = 'REPLACE INTO "%s" (key, value) VALUES (?, ?)' % self.tablename
        self.conn.executemany(UPDATE_ITEMS, items)
        if kwds:
            self.update(kwds)

    def __iter__(self):
        return self.iterkeys()

    def clear(self):
        if self.flag == 'r':
            raise RuntimeError('Refusing to clear read-only SqliteDict')

        # avoid VACUUM, as it gives "OperationalError: database schema has
        # changed"
        CLEAR_ALL = 'DELETE FROM "%s";' % self.tablename
        self.conn.commit()
        self.conn.execute(CLEAR_ALL)
        self.conn.commit()

    def commit(self, blocking=True):
        """
        Persist all data to disk.

        When `blocking` is False, the commit command is queued, but the data is
        not guaranteed persisted (default implication when autocommit=True).
        """
        if self.conn is not None:
            self.conn.commit(blocking)
    sync = commit

    def close(self, do_log=True, force=False):
        if do_log:
            logger.debug("closing %s" % self)
        if hasattr(self, 'conn') and self.conn is not None:
            if self.conn.autocommit and not force:
                # typically calls to commit are non-blocking when autocommit is
                # used.  However, we need to block on close() to ensure any
                # awaiting exceptions are handled and that all data is
                # persisted to disk before returning.
                self.conn.commit(blocking=True)
            self.conn.close(force=force)
            self.conn = None
        if self.in_temp:
            try:
                os.remove(self.filename)
            except:
                pass

    def terminate(self):
        """Delete the underlying database file. Use with care."""
        if self.flag == 'r':
            raise RuntimeError('Refusing to terminate read-only SqliteDict')

        self.close()

        if self.filename == ':memory:':
            return

        logger.info("deleting %s" % self.filename)
        try:
            if os.path.isfile(self.filename):
                os.remove(self.filename)
        except (OSError, IOError):
            logger.exception("failed to delete %s" % (self.filename))

    def __del__(self):
        # like close(), but assume globals are gone by now (do not log!)
        try:
            self.close(do_log=False, force=True)
        except Exception:
            # prevent error log flood in case of multiple SqliteDicts
            # closed after connection lost (exceptions are always ignored
            # in __del__ method.
            pass


# Adding extra methods for python 2 compatibility (at import time)
if major_version == 2:
    SqliteDict.__nonzero__ = SqliteDict.__bool__
    del SqliteDict.__bool__  # not needed and confusing
# endclass SqliteDict


class SqliteMultithread(Thread):
    """
    Wrap sqlite connection in a way that allows concurrent requests from multiple threads.

    This is done by internally queueing the requests and processing them sequentially
    in a separate thread (in the same order they arrived).

    """

    def __init__(self, filename, autocommit, journal_mode):
        super(SqliteMultithread, self).__init__()
        self.filename = filename
        self.autocommit = autocommit
        self.journal_mode = journal_mode
        # use request queue of unlimited size
        self.reqs = Queue()
        self.setDaemon(True)  # python2.5-compatible
        self.exception = None
        self.log = logging.getLogger('sqlitedict.SqliteMultithread')
        self.start()

    def run(self):
        if self.autocommit:
            conn = sqlite3.connect(
                self.filename, isolation_level=None, check_same_thread=False)
        else:
            conn = sqlite3.connect(self.filename, check_same_thread=False)
        conn.execute('PRAGMA journal_mode = %s' % self.journal_mode)
        conn.text_factory = str
        cursor = conn.cursor()
        conn.commit()
        cursor.execute('PRAGMA synchronous=OFF')

        res = None
        while True:
            req, arg, res, outer_stack = self.reqs.get()
            if req == '--close--':
                assert res, ('--close-- without return queue', res)
                break
            elif req == '--commit--':
                conn.commit()
                if res:
                    res.put('--no more--')
            else:
                try:
                    cursor.execute(req, arg)
                except Exception as err:
                    self.exception = (e_type, e_value, e_tb) = sys.exc_info()
                    inner_stack = traceback.extract_stack()

                    # An exception occurred in our thread, but we may not
                    # immediately able to throw it in our calling thread, if it has
                    # no return `res` queue: log as level ERROR both the inner and
                    # outer exception immediately.
                    #
                    # Any iteration of res.get() or any next call will detect the
                    # inner exception and re-raise it in the calling Thread; though
                    # it may be confusing to see an exception for an unrelated
                    # statement, an ERROR log statement from the 'sqlitedict.*'
                    # namespace contains the original outer stack location.
                    self.log.error('Inner exception:')
                    for item in traceback.format_list(inner_stack):
                        self.log.error(item)
                    # deliniate traceback & exception w/blank line
                    self.log.error('')
                    for item in traceback.format_exception_only(e_type, e_value):
                        self.log.error(item)

                    self.log.error('')  # exception & outer stack w/blank line
                    self.log.error('Outer stack:')
                    for item in traceback.format_list(outer_stack):
                        self.log.error(item)
                    self.log.error('Exception will be re-raised at next call.')

                if res:
                    for rec in cursor:
                        res.put(rec)
                    res.put('--no more--')

                if self.autocommit:
                    conn.commit()

        self.log.debug('received: %s, send: --no more--', req)
        conn.close()
        res.put('--no more--')

    def check_raise_error(self):
        """
        Check for and raise exception for any previous sqlite query.

        For the `execute*` family of method calls, such calls are non-blocking and any
        exception raised in the thread cannot be handled by the calling Thread (usually
        MainThread).  This method is called on `close`, and prior to any subsequent
        calls to the `execute*` methods to check for and raise an exception in a
        previous call to the MainThread.
        """
        if self.exception:
            e_type, e_value, e_tb = self.exception

            # clear self.exception, if the caller decides to handle such
            # exception, we should not repeatedly re-raise it.
            self.exception = None

            self.log.error('An exception occurred from a previous statement, view '
                           'the logging namespace "sqlitedict" for outer stack.')

            # The third argument to raise is the traceback object, and it is
            # substituted instead of the current location as the place where
            # the exception occurred, this is so that when using debuggers such
            # as `pdb', or simply evaluating the naturally raised traceback, we
            # retain the original (inner) location of where the exception
            # occurred.
            reraise(e_type, e_value, e_tb)

    def execute(self, req, arg=None, res=None):
        """
        `execute` calls are non-blocking: just queue up the request and return immediately.
        """
        self.check_raise_error()

        # NOTE: This might be a lot of information to pump into an input
        # queue, affecting performance.  I've also seen earlier versions of
        # jython take a severe performance impact for throwing exceptions
        # so often.
        stack = traceback.extract_stack()[:-1]
        self.reqs.put((req, arg or tuple(), res, stack))

    def executemany(self, req, items):
        for item in items:
            self.execute(req, item)
        self.check_raise_error()

    def select(self, req, arg=None):
        """
        Unlike sqlite's native select, this select doesn't handle iteration efficiently.

        The result of `select` starts filling up with values as soon as the
        request is dequeued, and although you can iterate over the result normally
        (`for res in self.select(): ...`), the entire result will be in memory.
        """
        res = Queue()  # results of the select will appear as items in this queue
        self.execute(req, arg, res)
        while True:
            rec = res.get()
            self.check_raise_error()
            if rec == '--no more--':
                break
            yield rec

    def select_one(self, req, arg=None):
        """Return only the first row of the SELECT, or None if there are no matching rows."""
        try:
            return next(iter(self.select(req, arg)))
        except StopIteration:
            return None

    def commit(self, blocking=True):
        if blocking:
            # by default, we await completion of commit() unless
            # blocking=False.  This ensures any available exceptions for any
            # previous statement are thrown before returning, and that the
            # data has actually persisted to disk!
            self.select_one('--commit--')
        else:
            # otherwise, we fire and forget as usual.
            self.execute('--commit--')

    def close(self, force=False):
        if force:
            # If a SqliteDict is being killed or garbage-collected, then select_one()
            # could hang forever because run() might already have exited and therefore
            # can't process the request. Instead, push the close command to the requests
            # queue directly. If run() is still alive, it will exit gracefully. If not,
            # then there's nothing we can do anyway.
            self.reqs.put(('--close--', None, Queue(), None))
        else:
            # we abuse 'select' to "iter" over a "--close--" statement so that we
            # can confirm the completion of close before joining the thread and
            # returning (by semaphore '--no more--'
            self.select_one('--close--')
            self.join()
# endclass SqliteMultithread



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provide extensive class and method for the built-in Exception system.
"""

import sys
import traceback


def get_last_exc_info():
    """Get last raised exception, and format the error message.
    """
    exc_type, exc_value, exc_tb = sys.exc_info()
    for filename, line_num, func_name, code in traceback.extract_tb(exc_tb):
        tmp = "{exc_value.__class__.__name__}: {exc_value}, appears in '{filename}' at line {line_num} in {func_name}(), code: {code}"
        info = tmp.format(
            exc_value=exc_value,
            filename=filename,
            line_num=line_num,
            func_name=func_name,
            code=code,
        )
        return info


class ExceptionHavingDefaultMessage(Exception):
    """A Exception class with default error message.
    """
    default_message = None

    def __str__(self):
        length = len(self.args)
        if length == 0:
            if self.default_message is None:
                raise NotImplementedError("default_message is not defined!")
            else:
                return self.default_message
        elif length == 1:
            return str(self.args[0])
        else:
            return str(self.args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.0.4"
__short_description__ = "Provide several commonly used logger."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"


try:
    from .logger import (
        DebugLevel,
        StreamOnlyLogger,
        SingleFileLogger,
        FileRotatingLogger,
        TimeRotatingLogger,
    )
    from .logfilter import find
except ImportError:  # pragma: no cover
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
from .scheduler import StatusFlag
from .scheduler_mongodb import MongoDBScheduler


class MongoDBStatusFlagScheduler(MongoDBScheduler, StatusFlag):
    """
    Similar to :class:`~pytq.scheduler_mongodb.MongoDBScheduler`.

    New Feature:

    There's pre-defined integer - ``duplicate_flag``, will be stored in
    ``status`` field. there's a ``edit_at`` datetime field, represent the
    last time the document been edited.

    .. note::

        Any value greater or equal than ``duplicate_flag``, AND the ``edit_at``
        time is smaller ``update_interval`` seconds ago, means it is a duplicate
        item.

    :param duplicate_flag: int, represent a status code for finished / duplicate
    :param update_interval: int, represent need-to-update interval (unit: seconds)
    """

    def __init__(self, logger=None, collection=None,
                 duplicate_flag=None, update_interval=None):
        super(MongoDBStatusFlagScheduler, self). \
            __init__(logger=logger, collection=collection)
        self.pre_process_duplicate_flag_and_update_interval(
            duplicate_flag, update_interval,
        )

    def _default_is_duplicate(self, task):
        """
        Check the status flag is greater or equal than
        :attr:`~StatusFlagScheduler.duplicate_flag` and last edit time is with
        in recent :attr:`~StatusFlagScheduler.update_interval` seconds.
        """
        n_sec_ago = datetime.utcnow() - timedelta(seconds=self.update_interval)
        filters = {
            "_id": task.id,
            self.status_key: {"$gte": self.duplicate_flag},
            self.edit_at_key: {
                "$gte": n_sec_ago,
            },
        }
        return self._col.find_one(filters) is not None

    def _get_finished_id_set(self):
        """
        It's Primary key value set.
        """
        n_sec_ago = datetime.utcnow() - timedelta(seconds=self.update_interval)
        filters = {
            self.status_key: {"$gte": self.duplicate_flag},
            self.edit_at_key: {
                "$gte": n_sec_ago,
            },
        }
        wanted = {"_id": True}
        return set([
            doc["_id"] for doc in self._col.find(filters, wanted)
        ])

    def _default_post_process(self, task):
        """
        Save output_data into ``out`` field.
        """
        self._col.update(
            {"_id": task.id},
            {
                "$set": {
                    self.output_key: self._encode(task.output_data),
                    self.status_key: self.duplicate_flag,
                    self.edit_at_key: datetime.utcnow(),
                },
            },
            upsert=True,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import string
import random
import logging
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

DEFAULT_LOG_FORMAT = "%(asctime)s; %(levelname)-8s; %(message)s"
DEFAULT_STREAM_FORMAT = "%(message)s"


def random_str(charset, length=32):
    """
    Generate random string.
    """
    return "".join([random.choice(charset) for _ in range(length)])


def get_logger_by_name(name=None):
    """
    Get a logger by name. If name is not provided, a random name will be used.
    """
    if name is None:
        name = random_str(string.hexdigits)
    logger = logging.getLogger(name)
    return logger


class DebugLevel(object):
    debug = logging.DEBUG
    info = logging.INFO
    warning = logging.WARNING
    error = logging.ERROR
    critical = logging.CRITICAL


def set_stream_handler(logger, stream_level, stream_format):
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(stream_level)
    stream_handler.setFormatter(logging.Formatter(stream_format))
    logger.addHandler(stream_handler)


class BaseLogger(object):
    """A base class for logger constructor.
    """
    tab = "    "
    enable_verbose = True
    _handler_cache = list()
    logger = None

    def debug(self, msg, indent=0):
        """Call logger.debug, indent format may apply.
        """
        self.logger.debug("%s%s" % (self.tab * indent, msg))

    def info(self, msg, indent=0):
        """Call logger.info, indent format may apply.
        """
        self.logger.info("%s%s" % (self.tab * indent, msg))

    def warning(self, msg, indent=0):
        """Call logger.warning, indent format may apply.
        """
        self.logger.warning("%s%s" % (self.tab * indent, msg))

    def error(self, msg, indent=0):
        """Call logger.error, indent format may apply.
        """
        self.logger.error("%s%s" % (self.tab * indent, msg))

    def critical(self, msg, indent=0):
        """Call logger.critical, indent format may apply.
        """
        self.logger.critical("%s%s" % (self.tab * indent, msg))

    def show(self, msg, indent=0):
        """Print message to console, indent format may apply.
        """
        if self.enable_verbose:
            sys.stderr.write("%s%s\n" % (self.tab * indent, msg))

    __call__ = show

    def remove_all_handler(self):
        """Unlink the file handler association.
        """
        for handler in self.logger.handlers[:]:
            self.logger.removeHandler(handler)
            self._handler_cache.append(handler)

    def recover_all_handler(self):
        """Relink the file handler association you just removed.
        """
        for handler in self._handler_cache:
            self.logger.addHandler(handler)
        self._handler_cache = list()


class StreamOnlyLogger(BaseLogger):
    """This logger only print message to console, and not write log to files.

    :param stream_format： log information format.

    **中文文档**

    只将日志打印到控制台, 并不将日志信息写入到文件。
    """

    def __init__(self, name=None,
                 stream_level=DebugLevel.info,
                 stream_format=DEFAULT_STREAM_FORMAT):
        logger = get_logger_by_name(name)

        # Set Logging Level
        logger.setLevel(logging.DEBUG)

        # Set Stream Handler
        set_stream_handler(logger, stream_level, stream_format)

        self.logger = logger


class SingleFileLogger(BaseLogger):
    """This logger print message to console and also write log to files.

    Only one log file will be used.

    :param reset: boolean, if True, the old log content will be removed.

    **中文文档**

    日志被写入到单个文件中。
    """

    def __init__(self, name=None, path=None,
                 logging_level=DebugLevel.debug,
                 stream_level=DebugLevel.info,
                 logging_format=DEFAULT_LOG_FORMAT,
                 stream_format=DEFAULT_STREAM_FORMAT,
                 reset=False,
                 ):
        logger = get_logger_by_name(name)
        if path is None:  # pragma: no cover
            raise ValueError("Please specify a file as log path!")

        # Set Logging Level
        logger.setLevel(logging_level)

        # Set File Handler
        if reset:
            with open(path, "wb") as f:
                pass

        file_handler = logging.FileHandler(
            path, mode="a", encoding="utf-8",
        )
        file_handler.setFormatter(logging.Formatter(logging_format))
        logger.addHandler(file_handler)

        # Set Stream Handler
        set_stream_handler(logger, stream_level, stream_format)

        self.logger = logger


class FileRotatingLogger(BaseLogger):
    """Definition:

    https://docs.python.org/2/library/logging.handlers.html#rotatingfilehandler

    :param max_bytes: max file size.
    :param backup_count: max number of files.

    **中文文档**

    当日志文件的体积大于某个阈值时, 自动重名名, 将日志录入到新的文件中。
    """

    def __init__(self, name=None, path=None,
                 logging_level=DebugLevel.debug,
                 stream_level=DebugLevel.info,
                 logging_format=DEFAULT_LOG_FORMAT,
                 stream_format=DEFAULT_STREAM_FORMAT,
                 max_bytes=1000000000,  # 1GB
                 backup_count=10,
                 ):
        logger = logging.getLogger(name)
        if path is None:  # pragma: no cover
            raise ValueError("Please specify a file as log path!")

        # Set Logging Level
        logger.setLevel(logging_level)

        # Set File Handler
        file_handler = RotatingFileHandler(
            path, mode="a",
            maxBytes=max_bytes, backupCount=backup_count,
            encoding="utf-8",
        )
        file_handler.setFormatter(logging.Formatter(logging_format))
        logger.addHandler(file_handler)

        # Set Stream Handler
        set_stream_handler(logger, stream_level, stream_format)

        self.logger = logger


class TimeRotatingLogger(BaseLogger):
    """Definition:

    https://docs.python.org/2/library/logging.handlers.html#timedrotatingfilehandler

    :param rotate_on_when: could be "h" (hour), "D" (Day).
    :param interval: rotate on how many hour/day.
    :param backup_count: max number of files.

    **中文文档**

    根据日志发生的时间, 每隔一定时间就更换一个文件名。
    """

    def __init__(self, name=None, path=None,
                 logging_level=DebugLevel.debug,
                 stream_level=DebugLevel.info,
                 logging_format=DEFAULT_LOG_FORMAT,
                 stream_format=DEFAULT_STREAM_FORMAT,
                 rotate_on_when="D",
                 interval=1,
                 backup_count=30,
                 ):
        logger = logging.getLogger(name)
        if path is None:  # pragma: no cover
            raise ValueError("Please specify a file as log path!")

        # Set Logging Level
        logger.setLevel(logging_level)

        # Set File Handler
        file_handler = TimedRotatingFileHandler(
            path,
            when=rotate_on_when,
            interval=interval,
            backupCount=backup_count,
            encoding="utf-8",
        )
        file_handler.setFormatter(logging.Formatter(logging_format))
        logger.addHandler(file_handler)

        # Set Stream Handler
        set_stream_handler(logger, stream_level, stream_format)

        self.logger = logger



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
If your log file is in a standard format, then this module can help you filter
out log information that you care aboue.

**中文文档**

如果你的日志格式是标准的 "%(asctime)s; %(levelname)-8s; %(message)s"。那么本
模块中的函数能够帮助您从日志中轻松的找到你感兴趣的结果。
"""

from __future__ import print_function


class Result(object):
    def __init__(self, path,
                 level, message, time_lower, time_upper, case_sensitive):
        self.path = path
        self.level = level
        self.message = message
        self.time_lower = time_lower
        self.time_upper = time_upper
        self.case_sensitive = case_sensitive

        self.lines = list()

    def __str__(self):
        return self.header + "\n" + "".join(self.lines)

    @property
    def header(self):
        template = ("--- Result of: filepath=%r, level=%r, pattern=%r,"
                    "time_lower=%r, time_upper=%r, case_sensitive=%r ---")
        return template % (self.path,
                           self.level, self.message,
                           self.time_lower, self.time_upper,
                           self.case_sensitive,
                           )

    def dump(self, path):
        with open(path, "wb") as f:
            f.write(str(self).encode("utf-8"))


def find(path,
         level=None,
         message=None,
         time_lower=None, time_upper=None,
         case_sensitive=False):  # pragma: no cover
    """
    Filter log message.

    **中文文档**

    根据level名称, message中的关键字, 和log的时间的区间, 筛选出相关的日志
    """
    if level:
        level = level.upper()  # level name has to be capitalized.

    if not case_sensitive:
        message = message.lower()

    with open(path, "r") as f:
        result = Result(path=path,
                        level=level, message=message,
                        time_lower=time_lower, time_upper=time_upper,
                        case_sensitive=case_sensitive,
                        )

        for line in f:
            try:
                _time, _level, _message = [i.strip() for i in line.split(";")]

                if level:
                    if _level != level:
                        continue

                if time_lower:
                    if _time < time_lower:
                        continue

                if time_upper:
                    if _time > time_upper:
                        continue

                if message:
                    if not case_sensitive:
                        _message = _message.lower()

                    if message not in _message:
                        continue

                result.lines.append(line)
            except Exception as e:
                print(e)

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is a minimized six model.
"""

import sys
import types

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str,
    integer_types = int,
    class_types = type,
    text_type = str
    binary_type = bytes

    MAXSIZE = sys.maxsize
else:
    string_types = basestring,
    integer_types = (int, long)
    class_types = (type, types.ClassType)
    text_type = unicode
    binary_type = str

    if sys.platform.startswith("java"):
        # Jython always uses 32 bits.
        MAXSIZE = int((1 << 31) - 1)
    else:
        # It's possible to have sizeof(long) != sizeof(Py_ssize_t).
        class X(object):

            def __len__(self):
                return 1 << 31
        try:
            len(X())
        except OverflowError:
            # 32-bit
            MAXSIZE = int((1 << 31) - 1)
        else:
            # 64-bit
            MAXSIZE = int((1 << 63) - 1)
        del X

if PY3:
    def iterkeys(d, **kw):
        return iter(d.keys(**kw))

    def itervalues(d, **kw):
        return iter(d.values(**kw))

    def iteritems(d, **kw):
        return iter(d.items(**kw))

else:
    def iterkeys(d, **kw):
        return d.iterkeys(**kw)

    def itervalues(d, **kw):
        return d.itervalues(**kw)

    def iteritems(d, **kw):
        return d.iteritems(**kw)


def add_metaclass(metaclass):
    """Class decorator for creating a class with a metaclass."""
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six
import attr

try:
    from .pkg.attrs_mate import AttrsClass
except:  # pragma: no cover
    from pytq.pkg.attrs_mate import AttrsClass


def none_or_is_callable(instance, attribute, value):
    """
    Can be None or callable.
    """
    if value is not None and not callable(value):
        raise TypeError(
            'callback must be a callable, got %s' % type(value).__name__)


@attr.s
class Task(AttrsClass):
    """
    Task is the core concept for task queue application.

    :param id: str or int, fingerprint of input_data.
    :param input_data: input data of the task.
    :param nth_counter: its the nth task in the entire queue.
    :param left_counter: there's nth task left for the entire batch job.
    :param output_data: after processing, the output data you got, could
        includes anything, such as, raw data, status, errors.
    :param pre_process: a callable function for single task, will be called
        before the process function been called.
    :param post_process: a callable function for single task, will be called
        after the process function been called.
    """
    id = attr.ib()
    input_data = attr.ib()
    nth_counter = attr.ib(default=None)  # integer
    left_counter = attr.ib(default=None)  # integer
    output_data = attr.ib(default=None)
    pre_process = attr.ib(  # a callable function
        default=None,
        validator=none_or_is_callable,
    )
    post_process = attr.ib(  # a callable function
        default=None,
        validator=none_or_is_callable,
    )

    def _pre_process(self):
        self.pre_process(self)

    def _post_process(self):
        self.post_process(self)

    def progress_msg(self):  # pragma: no cover
        """
        Generate progress message.
        """
        if self.nth_counter is None:
            msg = "Process: InputData(%r) ..." % self.input_data
            return msg

        if self.left_counter is None:
            msg = "Process %sth: InputData(%r) ..." % (
                self.nth_counter, self.input_data,
            )
        else:
            msg = "Process %sth: InputData(%r); %s left ..." % (
                self.nth_counter, self.input_data, self.left_counter,
            )
        return msg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from .pkg import sqlitedict
    from .scheduler import Task, BaseDBTableBackedScheduler, Encoder
except:  # pragma: no cover
    from pytq.pkg import sqlitedict
    from pytq.scheduler import Task, BaseDBTableBackedScheduler, Encoder


class SqliteDictScheduler(BaseDBTableBackedScheduler, Encoder):
    """
    A sqlite database backed scheduler. The primary key field is for
    fingerprint of input data, another field for output data.
    """
    user_db_path = None

    def __init__(self, logger=None, user_db_path=None):
        super(SqliteDictScheduler, self).__init__(logger=logger)

        self.link_encode_method()

        # initiate back end database
        self._dct = sqlitedict.SqliteDict(
            self.user_db_path, autocommit=True,
            encode=self._encode,
            decode=self._decode,
        )

    @property
    def user_db_path(self):
        """
        Back-end sqlite database file path.
        """
        raise NotImplementedError

    def _default_is_duplicate(self, task):
        """
        Check if ``task.id`` is presents in primary_key column.
        """
        return task.id in self._dct

    def _get_finished_id_set(self):
        """
        It's Primary key value set.
        """
        return set(self._dct.keys())

    def _default_post_process(self, task):
        """
        Write serialized output_data to another column.
        """
        self._dct[task.id] = task.output_data

    def __len__(self):
        return len(self._dct)

    def __iter__(self):
        return iter(self._dct)

    def clear_all(self):
        self._dct.clear()

    def get(self, id):
        return self._dct.get(id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.0.7"
__short_description__ = "A Task Queue Scheduler Framework."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"

try:
    from .task import Task
    from .scheduler import BaseScheduler, BaseDBTableBackedScheduler
    from .scheduler_sqlitedict import SqliteDictScheduler
    from .scheduler_sql import SqlScheduler
    from .scheduler_sql_status_flag import SqlStatusFlagScheduler
    from .scheduler_mongodb import MongoDBScheduler
    from .scheduler_mongodb_status_flag import MongoDBStatusFlagScheduler
    from .pkg import loggerFactory
except ImportError:  # pragma: no cover
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six
import sqlalchemy as sa
from datetime import datetime, timedelta
from .scheduler import StatusFlag
from .scheduler_sql import SqlScheduler


class SqlStatusFlagScheduler(SqlScheduler, StatusFlag):
    """
    Similar to :class:`~pytq.scheduler_sql.SqlScheduler`.

    New Feature:

    There's pre-defined integer - ``duplicate_flag``, will be stored in
    ``status`` column. there's a ``edit_at`` datetime column, represent the
    last time the row been edited.

    .. note::

        Any value greater or equal than ``duplicate_flag``, AND the ``edit_at``
        time is smaller ``update_interval`` seconds ago, means it is a duplicate
        item.

    :param duplicate_flag: int, represent a status code for finished / duplicate
    :param update_interval: int, represent need-to-update interval (unit: seconds)
    """

    def __init__(self, logger=None, uri=None, table=None,
                 duplicate_flag=None, update_interval=None):
        super(SqlStatusFlagScheduler, self). \
            __init__(logger=logger, uri=uri, table=table)
        self.pre_process_duplicate_flag_and_update_interval(
            duplicate_flag, update_interval,
        )

    def prepare_table(self, table):
        if table is not None:
            self.table = table
        if self.table is None:
            raise NotImplementedError(
                "Please specify table name or sqlalchemy.Table!")

        if isinstance(self.table, sa.Table):
            metadata = self.table.metadata
            metadata.create_all(self.engine)
        elif isinstance(self.table, six.string_types):
            metadata = sa.MetaData()
            if self.id_type_is_str:
                id_column = sa.Column(self.id_key, sa.String(), primary_key=True)
            else:
                id_column = sa.Column(self.id_key, sa.Integer(), primary_key=True)
            table = sa.Table(
                self.table, metadata,
                id_column,
                sa.Column(self.out_key, sa.PickleType()),
                sa.Column(self.status_key, sa.Integer()),
                sa.Column(self.edit_at_key, sa.DateTime())
            )
            metadata.create_all(self.engine)
            self.table = table

    @property
    def status_col(self):
        return getattr(self.table.c, self.status_key)

    @property
    def edit_at_col(self):
        return getattr(self.table.c, self.edit_at_key)

    def _default_is_duplicate(self, task):
        """
        Check the status flag is greater or equal than
        :attr:`~StatusFlagScheduler.duplicate_flag` and last edit time is with
        in recent :attr:`~StatusFlagScheduler.update_interval` seconds.
        """
        n_sec_ago = datetime.utcnow() - timedelta(seconds=self.update_interval)
        sql = sa.select(
            [sa.func.count(self.id_col), ]
        ).where(
            sa.and_(
                self.id_col == task.id,
                self.status_col >= self.duplicate_flag,
                self.edit_at_col >= n_sec_ago,
            )
        )
        row = self.connection.execute(sql).fetchone()
        if row[0] == 1:
            return True
        else:
            return False

    def _get_finished_id_set(self):
        """
        It's Primary key value set.
        """
        n_sec_ago = datetime.utcnow() - timedelta(seconds=self.update_interval)
        sql = sa.select([self.id_col]).\
            where(
                sa.and_(
                    self.status_col >= self.duplicate_flag,
                    self.edit_at_col >= n_sec_ago,
                )
        )
        cursor = self.connection.execute(sql)
        return set([row[0] for row in cursor])

    def _default_post_process(self, task):
        """
        Write serialized output_data to "_out" column.
        """
        ins = self.table.insert()
        row = {
            self.id_key: task.id,
            self.out_key: task.output_data,
            self.status_key: self.duplicate_flag,
            self.edit_at_key: datetime.utcnow(),
        }
        try:
            self.connection.execute(ins, row)
        except sa.exc.IntegrityError:
            row.pop(self.id_key)
            upd = self.table.update(). \
                where(self.id_col == task.id). \
                values(**row)
            self.connection.execute(upd)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six
import sqlalchemy as sa
from multiprocessing import cpu_count

try:
    from .scheduler import Task, BaseDBTableBackedScheduler, Encoder
except:  # pragma: no cover
    from pytq.scheduler import Task, BaseDBTableBackedScheduler, Encoder


class SqlScheduler(BaseDBTableBackedScheduler, Encoder):
    """
    A SQL database backed scheduler. Fingerprint of input data will be stored
    under :attr:`~SqlScheduler.id_key` column ("_id" by default). output data
    will be serialized and stored under :attr:`~SqlScheduler.out_key` column.

    :param uri: connection string for sqlaclhemy.create_engine(uri)
        Example: postgresql://username:password@localhost:5432/mydatabase
    :param table: str or sa.Table. If you use the default post process,
        the table has to have `_out` column. But a string name is recommended.
    """
    uri = None
    table = None

    id_key = "_id"
    """id column name
    """

    out_key = "_out"
    """output column name
    """

    id_type_is_str = True

    def __init__(self, logger=None, uri=None, table=None):
        super(SqlScheduler, self).__init__(logger=logger)
        self.link_encode_method()
        self.prepare_uri(uri)

        n_cpu = cpu_count()
        if n_cpu >= 8:
            pool_size = n_cpu
        else:
            pool_size = 5
        self.engine = sa.create_engine(self.uri, pool_size=pool_size)
        self.connection = self.engine.connect()
        self.prepare_table(table)

    def prepare_uri(self, uri):
        if uri is not None:
            self.uri = uri
        if self.uri is None:
            raise NotImplementedError("Please specify db uri connect info!")

    def prepare_table(self, table):
        if table is not None:
            self.table = table
        if self.table is None:
            raise NotImplementedError(
                "Please specify table name or sqlalchemy.Table!")

        if isinstance(self.table, sa.Table):
            metadata = self.table.metadata
            metadata.create_all(self.engine)
        elif isinstance(self.table, six.string_types):
            metadata = sa.MetaData()
            if self.id_type_is_str:
                id_column = sa.Column(self.id_key, sa.String(), primary_key=True)
            else:
                id_column = sa.Column(self.id_key, sa.Integer(), primary_key=True)
            table = sa.Table(
                self.table, metadata,
                id_column,
                sa.Column(self.out_key, sa.PickleType()),
            )
            metadata.create_all(self.engine)
            self.table = table

    @property
    def t(self):
        return self.table

    @property
    def id_col(self):
        return getattr(self.table.c, self.id_key)

    @property
    def out_col(self):
        return getattr(self.table.c, self.out_key)

    def __enter__(self):
        try:
            self.connection.close()
        except:
            pass
        self.connection = self.engine.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()

    def connect(self):
        try:
            self.connection.close()
        except:
            pass
        self.connection = self.engine.connect()

    def close(self):
        self.connection.close()

    def _default_is_duplicate(self, task):
        """
        Check if ``task.id`` is presents in primary_key column.
        """
        sql = sa.select(
            [sa.func.count(self.id_col), ]
        ).where(self.id_col == task.id)
        row = self.connection.execute(sql).fetchone()
        if row[0] == 1:
            return True
        else:
            return False

    def _get_finished_id_set(self):
        """
        It's Primary key value set.
        """
        sql = sa.select([self.id_col])
        cursor = self.connection.execute(sql)
        return set([row[0] for row in cursor])

    def _default_post_process(self, task):
        """
        Write serialized output_data to "_out" column.
        """
        ins = self.table.insert()
        row = {
            self.id_key: task.id,
            self.out_key: task.output_data,
        }
        try:
            self.connection.execute(ins, row)
        except sa.exc.IntegrityError:
            row.pop(self.id_key)
            upd = self.table.update(). \
                where(self.id_col == task.id). \
                values(**row)
            self.connection.execute(upd)

    def __len__(self):
        sql = sa.select([sa.func.count(self.id_col), ])
        return self.connection.execute(sql).fetchone()[0]

    def __iter__(self):
        sql = sa.select([self.id_col])
        for row in self.connection.execute(sql):
            yield row[0]

    def clear_all(self):
        self.connection.execute(self.table.delete())

    def get(self, id):
        sql = sa.select([self.out_col]). \
            where(self.id_col == id)
        row = self.connection.execute(sql).fetchone()
        return row[0]
